# Generated by Django 4.2.2 on 2023-08-06 09:46

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lnschema_core", "0013_remove_feature_labels_orm_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="featureset",
            old_name="ref_field",
            new_name="registry",
        ),
    ]
